{% extends '!footer.html' %}
{% block extrafooter %}
<br><br>
<p>The Observatory Platform documentation is licensed under a
<a href="https://creativecommons.org/licenses/by/4.0/" target="_blank">Creative Commons Attribution 4.0 International
(CC BY 4.0)</a> license, with the exception of the documentation's code samples, which are licensed under a
<a href="https://creativecommons.org/publicdomain/zero/1.0/deed.en" target="_blank">CC0 1.0 Universal (CC0 1.0) Public
Domain Dedication</a>.</p>
{% endblock %}