{# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Richard Hosking #}
{#
The purpose of this script it to export the unique book list from the book_product table
Primarily, the goal is to create a flat structure which is suitable for graphing in Kibana
#}

SELECT
    ISBN13 as product_id,
    work_id,
    work_family_id,
    onix.ProductForm,
    (ARRAY_LENGTH(months) > 0) as usage_flag,
    onix.EditionNumber,
    CAST(onix.published_year as INT64) as published_year,
    onix.published_date AS published_date,
    onix.title,
    onix.bic_subjects,
    onix.bisac_subjects,
    onix.thema_subjects,
    onix.keywords,
    onix.authors
FROM `{{ book_product_table_id }}`